//
//  ParalellMA.c
//  
//
//  Created by Lucas Angelo on 28/09/15.
//
//

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include "ParalellMA.h"
#include <math.h>
#include "structure.h"
#include "../Fitness/Processing.c"
#include "Calc_entropy.h"


#define WORKTAG 1
#define TAGLS 3
#define DIETAG 2
#define DEBUG 0
#define TAM_BUFFER 1000

int * identity;
int size;
int beta;
void read_permutation_of_archivo(char* name, permutation *perm)
{
    int i,n;
    FILE *filePointer;
    if ((filePointer = fopen(name,"r")) == NULL){
        printf("Atencao arquivo nao encontrado\n");
    }
    else
    {
        //read size of permutation
        fscanf(filePointer,"%d",&n);
        perm->size = n;
        size=n;
        //alocate memory for perm->pi
        perm->pi = malloc(perm->size*sizeof(int));
        //leer permutacion
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
        fscanf(filePointer,"%d",&n);
        identity= (int *) calloc (n, sizeof (int));
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&identity[i]);
        }
        
        
    }//end  else
    fclose(filePointer);
}//end read permutation of archivo


void population_init(population *pob, permutation *perm, float porc_reemp, float porc_selection){
	int i,j,r,n,prob;
	
    //Calc size of population
	n = perm->size;
    //printf("n%d\n",n);
	pob->sizePopulation =  (int) (n * (log(n) / log(2))); //siz of population = (nlogn); log in base 2
	pob->basePopulation = 0;
	pob->topPopulation = pob->sizePopulation;
    
	pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de population what go to crossover
	pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a partir de where will replace
    
	pob->baseDescendencia = pob->sizePopulation;
	pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
	pob->sizeTotal = pob->topeDescendencia;// tamanho total = poblacion + descendencia
    
    /*alocar memoria*/
	pob->genome = malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in population_init(x,x)\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = malloc(n * sizeof(int));//alocar memoria para cada genome
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in population_init(x,x)\n");
    }
    
	/*create population a partir of  permutacion*/
    // A genome is see with a permutation of concatenation chroms
    
	for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        
		for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//valores entre 0 y 1
                if (prob == 1)//si prob 1, entonces gen positivo
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                }
                else//si prob 0, entonces gen negativo
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                    
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
                
            }
        }//end for interno
        
		pob->genome[i].size = perm->size;
        //eliminar_repetidas(&pob,pob->sizePopulation,adicional,n);
	}//end for extern
	/*agregar tamanho a desc*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
		pob->genome[i].size = perm->size;
	}
    
    
}//end create init population

void countingSort(population *pob, int k){
	int i,j;
	int *c;//c[k];
	Genome *a;
    
    /*alocar memoria para c[k]*/
    c = malloc(k*sizeof(int));
    
	a = malloc(pob->sizePopulation*sizeof(Genome));
	if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
	for(i=0; i<pob->sizePopulation; i++){
		a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
	}
    
	/*counting sort*/
	for(i=0; i<k; i++){
		c[i] = 0;
	}
    
	for(i=0; i<pob->sizePopulation; i++){
		c[a[i].fitness] = c[a[i].fitness] + 1;
	}
    
	for(i=1; i<k; i++){
		c[i] = c[i] + c[i-1];
	}
    
	for(i=pob->sizePopulation-1; i>=0; i--){
		
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
		c[a[i].fitness] = c[a[i].fitness] - 1;
	}
    
    
    /*Liberar memoria*/
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
	free(a);
    free(c);
}
void crossover(population *pob, permutation *perm,float prob_crossover){
	int i,j,k,pos_pai1, pos_pai2, point_crossover;
	float prob;
	k = pob->baseDescendencia;//indice para colocar en descendencia
	/*crossver sobre populacao atual*/
	for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
		//se falta ultima posicao para colocar descendencia, terminar
		if (k+1 == pob->sizeTotal)
			break;
        
		//escolher posicao dos pais
		pos_pai1 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		//buscar pos2 que no seja igual a pos1, pelo  menos 10vezes
		for(j=0; j<10; j++){
			if (pos_pai1 != pos_pai2)
                break;
			pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		}
        
		prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
		if (prob <= prob_crossover)
        {//crossover
			point_crossover = rand() % perm->size;//valores entre 0 e tamanho de permutation - 1
			//aplicar single_point_crossover na primera parte antes do ponto de crossover
			for(j=0; j<point_crossover; j++){
				pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			//aplicar single_point_crossover na segunda parte depois do ponto de crossover
			for(j=point_crossover; j<perm->size; j++){
				pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			k = k + 2;
		}//end if crossover
	}//end for iniit
	pob->topeDescendencia = k;//atualizar tope descendencia
}

//operator of selection, the ideia is sort population using the fitnes of each indidividual
void selection(population *pob,  int n_generetion, int * best_solution){
  
   // printf("entrei no selection\n");
    countingSort(pob,pob->genome[0].size+2);
 //   printf("ordenei\n");

    if (pob->genome[0].fitness < *best_solution)
        *best_solution = pob->genome[0].fitness;
    
 //   printf("meu melhor e %d finalizei selection\n", *best_solution);

}//end selection



void mutation(population *pob, permutation *perm, float prob_mutation){
	int i,j;
	float prob;
    
	/*mutation about descendents, baseDescendencia=0*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		for(j=0; j<perm->size; j++)
        {
			prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
			
            if(prob <= prob_mutation)
            {//applied a gene mutation (flip sign)
                
                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
                
            }
            
        }
        
    }//end for
    
}//end of the function mutation

void replacement(population *pob){
	int i,j,pos_reemp;
    int conta=0;
	/*substituir descendentes na populacao atual*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		//buscar pos de substituicao
		int tamanhoReemp = pob->topPopulation - pob->baseReplace;
		pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(valores entre 0 e tamanhoReemp-1)+baseReplace
		
        //realizar substituicao
		if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
			//copiar  fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copiar genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
	}
}//end replacement



/* operator local search based in mutation*/
/* The operator is applied over one individual "i"
 * Return 1(true) when we have better results of the indidual i"
 * otherwise return 0(false) */
int auxlocal_search2(population *pob, int i){
	int k, r, replace, fitness, pos_gen, better;
	k=0;
	replace = 0;//false
	better = pob->genome[i].fitness;
    pos_gen=0;
	while(k<2)
    {
		while(1 && (qtd_genes/qtd_cromossomo>2))
		{
            
            r=rand();
            pos_gen = r % pob->genome[0].size;//values between 0 and (size)-1
            if((pos_gen!=0) && (pos_gen != pob->genome[0].size-1) && (pob->genome[i].genes[pos_gen-1]!=0)&&
               (pob->genome[i].genes[pos_gen+1]!=0) && (pob->genome[i].genes[pos_gen]!=0))
                break;
        }//end while
        
        //modificar gen en "pos_gen"
        if (pos_gen)
        {
            
            
            pob->genome[i].genes[pos_gen] = pob->genome[i].genes[pos_gen]*(-1);

            
	        load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
	        fitness =distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            if (fitness < better){
                pob->genome[i].fitness = fitness;
                replace = 1;//true
                break;
                
            }
            else{
                pob->genome[i].genes[pos_gen] =pob->genome[i].genes[pos_gen]*(-1);//restore previous state
            }
        }
        k++;
	}//end while
    
	return replace;
}



int * auxlocal_search(msgGenome input){
	int k, r, i, fitness, pos_gen, better, flag;
	k=0;
    flag=0;
    int * replace=calloc (3, sizeof (int));// posicao 0 o indice pra alterar e posicao 1 o fitness e posicao 2 o indice na populacao
	better = input.fitness;
    pos_gen=input.pos1;
    int NumGene=input.size_msg;
    int *genomaA=(int *) calloc (NumGene, sizeof (int));
    int *genomaB=(int *) calloc (NumGene, sizeof (int));
    
    
    while (k<2)
    {
        
        if (pos_gen)
        {
            if (k==0)
            {
                for(i=0; i<NumGene; i++){
                    if (pos_gen==i) {
                        genomaA[i]=input.genes[i]*-1;
                    }
                    else
                        genomaA[i]=input.genes[i];
                    
                    genomaB[i]=input.identity[i];
                }
            }
            
            else
                genomaA[pos_gen]=genomaA[pos_gen]*-1;
            
            
            
            
            load_genoma( genomaA, genomaB,NumGene, NumGene);
            fitness = distance(genomaA,genomaB, NumGene,2);
            
            
            
            
            if (fitness < better){
                replace[0]=pos_gen;
                replace[1]=fitness;
                replace[2]=input.work_number;
                flag=1;
                break;
                
            }
            else{
                genomaA[pos_gen]=genomaA[pos_gen]*-1;//restore previous state
            }
            
            
        }//end if do pos_gen
    
        pos_gen=input.pos2;
        k++;
    
    }//while
        
        if (!flag) {
            replace[0]=-1;//nao houve melhoras
            replace[1]=0;
            replace[2]=0;
        }
    
    if (input.work_number==328) {
        int m;
        for (m=0; m<NumGene; m++ ) {
            printf("%d ", genomaA[m]);
        }
        printf("\n");
        
        printf("distance %d\n", replace[1]);
    }
    
    //printf("---%d\n",replace[2]);
   /*
    while(k<2)
    {
		while(1 && (qtd_genes/qtd_cromossomo>2))
		{
    
            r=rand();
            pos_gen = r % input.size_msg;//values between 0 and (size)-1
            if((pos_gen!=0) && (pos_gen != (input.size_msg-1)) && (input.genes[pos_gen-1]!=0)&&
               (input.genes[pos_gen+1]!=0) && (input.genes[pos_gen]!=0))
                break;
        }//end while
        
        //modificar gen en "pos_gen"
        if (pos_gen)
        {
            if (k==0)
            {
                for(i=0; i<NumGene; i++){
                    if (pos_gen==i) {
                        genomaA[i]=input.genes[i]*-1;
                    }
                    else
                        genomaA[i]=input.genes[i];
                    
                    genomaB[i]=input.identity[i];
                }
            }
            
            else
                 genomaA[pos_gen]=genomaA[pos_gen]*-1;
           
            load_genoma( genomaA, genomaB,NumGene, NumGene);
            fitness = distance(genomaA,genomaB, NumGene,2);
            
            if (fitness < better){
                replace[0]=pos_gen;
                replace[1]=fitness;
                 replace[2]=input.work_number;
                flag=1;
                break;
                
            }
            else{
                genomaA[pos_gen]=genomaA[pos_gen]*-1;//restore previous state
            }
            
            
        }//end if do pos_gen
        k++;
	}//end while
    free(genomaA);
    free(genomaB);
    
    if (!flag) {
        replace[0]=-1;//nao houve melhoras
        replace[1]=0;
        replace[2]=0;
    }
    */
    
	return replace;
}

/* Procedure that applies Local Search over the population
 * if nro_generacion is equal to 1, it means that is the initial pop0
 */
void local_search(population *pob, int nro_generation,float probability_ls, float porc_localsearch, int *sementes){
    int i,proc, myid, numprocs,work_number,limit_search;
    float randomico;
	msgBusca input_buffer;
    msgGenome output_buffer;
    proc = 1;
    int contador_m=0;
    MPI_Status status;
    int pop0=0;
    int t1;
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    
    work_number = pob->basePopulation;
    // printf("entrei no local_search\n");
    if (nro_generation > 1)
    {

        limit_search = pob->sizeTotal * porc_localsearch;
        
        
            //Seed the slaves; send one unit of work to each slave.
            if (DEBUG) printf("... Seed the slaves; send one unit of work to each slave.\n");
        
            while(proc < numprocs)
            {
            if (work_number == limit_search)
                break;
            
            output_buffer.work_number = work_number;//indice
            output_buffer.size_msg = pob->genome[work_number].size;
            for(i=0; i<output_buffer.size_msg; i++){
                output_buffer.genes[i] = pob->genome[work_number].genes[i];
                output_buffer.identity[i] = identity[i];
            }
            output_buffer.flag=0;
            output_buffer.fitness=pob->genome[work_number].fitness;
                scanf("%d",&sementes[beta]);
            output_buffer.pos1=sementes[beta];
            // printf("%d\n", sementes[beta]);
            beta++;
                scanf("%d",&sementes[beta]);
            output_buffer.pos2=sementes[beta];
            //printf("%d\n", sementes[beta]);
            beta++;
            MPI_Send(&output_buffer,
                     sizeof(msgGenome),
                     MPI_BYTE,
                     proc, //destination process
                     TAGLS,
                     MPI_COMM_WORLD);
            work_number++;
            proc++;
        }//for
        
        
        
        
        
    // Loop over getting new work requests until there is no more work
            //to be done
            if (DEBUG) printf("... Loop over getting new work requests\n");
        
            while (work_number < limit_search)
            {
            //Receive results from a slave
            MPI_Recv(&input_buffer,
                     sizeof(msgBusca),
                     MPI_BYTE,
                     MPI_ANY_SOURCE, //from any slave ID
                     TAGLS,
                     MPI_COMM_WORLD,
                     &status);
            
            if (input_buffer.index_gene!=-1 && input_buffer.result!=0)
            {
                
                pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
                pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene] = pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene]*-1;
            }//end if
            
            //Send the slave a new work unit
            output_buffer.work_number = work_number;
            output_buffer.size_msg = pob->genome[work_number].size;
            for(i=0; i<output_buffer.size_msg; i++){
                output_buffer.genes[i] = pob->genome[work_number].genes[i];
                output_buffer.identity[i] = identity[i];
            }
            output_buffer.flag=0;
            output_buffer.fitness=pob->genome[work_number].fitness;
                scanf("%d",&sementes[beta]);
                output_buffer.pos1=sementes[beta];
            //  printf("%d\n", sementes[beta]);
                beta++;
                scanf("%d",&sementes[beta]);
                output_buffer.pos2=sementes[beta];
              // printf("%d\n",sementes[beta]);
                beta++;

            MPI_Send(&output_buffer,
                     sizeof(msgGenome),
                     MPI_BYTE,
                     status.MPI_SOURCE, //to slave ID, who sent last message
                     TAGLS,
                     MPI_COMM_WORLD);
            
            work_number++;
        }//end while
        
        
            if (work_number==limit_search)
            {
            
            //There's no more work to be done, so receive all the outstanding
            //results from the slaves.
            if (DEBUG) printf("... There's no more work to be done, so receive all\n");
            for (proc = 1; proc<numprocs; proc++)
            {
                MPI_Recv(&input_buffer,
                         sizeof(msgBusca),
                         MPI_BYTE,
                         MPI_ANY_SOURCE, //from any slave ID
                         TAGLS,
                         MPI_COMM_WORLD,
                         &status);
                
                if (input_buffer.index_gene!=-1 && input_buffer.result!=0)
                {
                    
                    pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
                    pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene] = pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene]*-1;
                }//end if
                
            }//end for
            
        }//end if
        
        
    
    }//end if
    else
    {
        //nro_generacion == 1, we have initial population
        
        limit_search=pob->topPopulation;
        
            //Seed the slaves; send one unit of work to each slave.
            if (DEBUG) printf("... Seed the slaves; send one unit of work to each slave.\n");
        while(proc < numprocs)
        {
            if (work_number == limit_search)
                break;
            
            output_buffer.work_number = work_number;//indice
            output_buffer.size_msg = pob->genome[work_number].size;
            for(i=0; i<output_buffer.size_msg; i++){
                output_buffer.genes[i] = pob->genome[work_number].genes[i];
                output_buffer.identity[i] = identity[i];
            }
            output_buffer.flag=0;
            output_buffer.fitness=pob->genome[work_number].fitness;
            
            scanf("%d",&sementes[beta]);
            //printf("%d\n", sementes[beta]);
            output_buffer.pos1=sementes[beta];
            beta++;
           
            scanf("%d",&sementes[beta]);
          //  printf("%d\n", sementes[beta]);
            output_buffer.pos2=sementes[beta];
            beta++;
            if (work_number==0) {
                
            }
           // printf(" Send %d\n",work_number);
            contador_m++;
            MPI_Send(&output_buffer,
                     sizeof(msgGenome),
                     MPI_BYTE,
                     proc, //destination process
                     TAGLS,
                     MPI_COMM_WORLD);
            work_number++;
            proc++;
        }//for
        
        
    
            
        // Loop over getting new work requests until there is no more work
        //to be done
        if (DEBUG) printf("... Loop over getting new work requests\n");
            
        while (work_number < limit_search)
        {
            //Receive results from a slave
            MPI_Recv(&input_buffer,
                        sizeof(msgBusca),
                        MPI_BYTE,
                        MPI_ANY_SOURCE, //from any slave ID
                        TAGLS,
                        MPI_COMM_WORLD,
                        &status);
            
            contador_m++;
            
            
            // printf("%d - Rec %d \n",work_number,input_buffer.index_pop);
            
            if (input_buffer.index_pop==0) {
                
               // printf(" Estou recebendo2 Pop %d recebi resultado zero ??? %d %d\n",input_buffer.index_pop,input_buffer.result,MPI_ANY_SOURCE);
                
            }
            
            if (input_buffer.index_gene!=-1 && input_buffer.result!=0 )
            {
                pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
                pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene] = pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene]*-1;
            }//end if
                
            //Send the slave a new work unit
            output_buffer.work_number = work_number;
            output_buffer.size_msg = pob->genome[work_number].size;
            for(i=0; i<output_buffer.size_msg; i++){
                output_buffer.genes[i] = pob->genome[work_number].genes[i];
                output_buffer.identity[i] = identity[i];
            }
            output_buffer.flag=0;
            output_buffer.fitness=pob->genome[work_number].fitness;
            scanf("%d",&sementes[beta]);
           // printf("%d\n", sementes[beta]);
            output_buffer.pos1=sementes[beta];
            beta++;
            
            
            scanf("%d",&sementes[beta]);
            output_buffer.pos2=sementes[beta];
            
            // printf("%d\n", sementes[beta]);
            beta++;
            
            MPI_Send(&output_buffer,
                        sizeof(msgGenome),
                        MPI_BYTE,
                        status.MPI_SOURCE, //to slave ID, who sent last message
                        TAGLS,
                        MPI_COMM_WORLD);
                
            work_number++;
            fflush(stdout);
        }//end while
        
     //   printf("contador_m %d\n",contador_m);
        if (work_number==limit_search)
        {
                
            //There's no more work to be done, so receive all the outstanding
            //results from the slaves.
            if (DEBUG) printf("... There's no more work to be done, so receive all\n");
            for (proc = 1; proc<numprocs; proc++)
            {
                MPI_Recv(&input_buffer,
                            sizeof(msgBusca),
                            MPI_BYTE,
                            MPI_ANY_SOURCE, //from any slave ID
                            TAGLS,
                            MPI_COMM_WORLD,
                            &status);
                
                //  printf("- Rec %d \n",input_buffer.index_pop);
                if (input_buffer.index_pop==0 ) {
                    
                    //printf(" Estou recebendo3 Pop %d recebi resultado zero ??? %d \n",input_buffer.index_pop,input_buffer.result);
                    
                }
                if (input_buffer.index_gene!=-1&& input_buffer.result!=0 )
                {
                        
                    pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
                    pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene] = pob->genome[input_buffer.index_pop].genes[input_buffer.index_gene]*-1;
                }//end if
                
            }//end for
                
        }//end if
        
        
    }//end else
}



/* method that resert the population, conservando una parte
 * e gerando aleatoriamente o resto
 */
void restarting_population(population *pob, permutation *perm, float porc_retain){
	
    int i,j, prob,r, size_conservation,myid, numprocs, tope;
    msgBusca input_buffer;
    msgGenome output_buffer;
    int work_number;
    MPI_Status status;
    
	// retain a portion of the population
	size_conservation = pob->sizePopulation * porc_retain;
    
	for(i=size_conservation; i < pob->sizePopulation; i++)
    {
		for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//valores entre 0 y 1
                if (prob == 1)
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                    
                }
                else//si prob 0, entonces gen negativo
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
            }
        }//end for interno
        
		pob->genome[i].size = perm->size;
        
	}//end for
    
    
    
    
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    
    work_number = size_conservation;
    tope=pob->sizePopulation;
    
    
    //Seed the slaves; send one unit of work to each slave.
    if (DEBUG) printf("... Seed the slaves; send one unit of work to each slave.\n");
    int proc;//proc 0 is the MASTER!
    for (proc = 1; proc<numprocs; proc++)
    {
        if (work_number == tope)
            break;
        
        output_buffer.work_number = work_number;
        output_buffer.size_msg = pob->genome[work_number].size;
        for(i=0; i<output_buffer.size_msg; i++)
        {
            output_buffer.genes[i] = pob->genome[work_number].genes[i];
            output_buffer.identity[i] = identity[i];
        }
        
        MPI_Send(&output_buffer,
                 sizeof(msgGenome),
                 MPI_BYTE,
                 proc, //destination process
                 WORKTAG,
                 MPI_COMM_WORLD);
        
        work_number++;
        
    }//end for
    
    
    // Loop over getting new work requests until there is no more work
    //to be done
    if (DEBUG) printf("... Loop over getting new work requests\n");
    
    while (work_number < tope)
    {
        //Receive results from a slave
        MPI_Recv(&input_buffer,
                 sizeof(msgBusca),
                 MPI_BYTE,
                 MPI_ANY_SOURCE, //from any slave ID
                 MPI_ANY_TAG,
                 MPI_COMM_WORLD,
                 &status);
        pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
        auxlocal_search2(pob, input_buffer.index_pop);
        //Send the slave a new work unit
        output_buffer.work_number = work_number;
        output_buffer.size_msg = pob->genome[work_number].size;
        for(i=0; i<output_buffer.size_msg; i++){
            output_buffer.genes[i] = pob->genome[work_number].genes[i];
        }
        
        MPI_Send(&output_buffer,
                 sizeof(msgGenome),
                 MPI_BYTE,
                 status.MPI_SOURCE, //to slave ID, who sent last message
                 WORKTAG,
                 MPI_COMM_WORLD);
        
        //if (DEBUG) printf("processor MASTER %d  sent %d\n",myid,output_buffer);
        
        work_number++;
    }//end while
    
    //There's no more work to be done, so receive all the outstanding
    //results from the slaves.
    if (DEBUG) printf("... There's no more work to be done, so receive all\n");
    for (proc = 1; proc<numprocs; proc++)
    {
        MPI_Recv(&input_buffer,
                 sizeof(msgBusca),
                 MPI_BYTE,
                 MPI_ANY_SOURCE, //from any slave ID
                 MPI_ANY_TAG,
                 MPI_COMM_WORLD,
                 &status);
        pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
        auxlocal_search2(pob, input_buffer.index_pop);
    }//end for
    
    
    

    
}//end restarting_population

//this function return 1 for reset population and 0 otherwise
int reset_population(population *pob, int i, float limiteentropia){
    float value=entropia_shannon(pob);
    //printf("estou no reset population %f < %f\n", value, limiteentropia);
    
    if ( value < limiteentropia)
    {
        return 1;
        
    }
    
    else
        return 0;
    
}

void see_population(permutation *perm,population *pob)
{
    int i,j;
    printf("mostrando a populacao gerada em forma de permutacao\n");
    for(i=0; i<pob->sizePopulation;i++ )
    {
        printf("%d ",i);
        for(j=0;j<perm->size;j++)
        {
            printf("%d,", pob->genome[i].genes[j]);
            
        }
        printf("\n");
    }
}//end see_population



void see_fitness(population *pob ,int flag)
{
    int i;
    if(!flag)
    {
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
            printf("%d: distance is:%d\n",i, pob->genome[i].fitness);
    }
    else
    {
        
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
            printf("distance is:%d\n",pob->genome[i].fitness);
    }
    
}


int calcular_fitness_mpi(msgGenome input)
{
    int NumGene=input.size_msg;
    int i;
   // printf("tamanho %d\n", NumGene);
    
    //alocar genomas
    
    int *genomaA=(int *) calloc (NumGene, sizeof (int));
    int *genomaB=(int *) calloc (NumGene, sizeof (int));
    
    for(i=0; i<NumGene; i++){
        genomaA[i]=input.genes[i];
        genomaB[i]=input.identity[i];
	}
    
    load_genoma( genomaA, genomaB,NumGene, NumGene);
    int valor = distance(genomaA,genomaB, NumGene,2);
    
    free(genomaA);
    free(genomaB);
    return valor;
    
    
}//end function


void fitness_mpi(population *pob, int base, int tope){
    int i,myid, numprocs;
    msgBusca input_buffer;
    msgGenome output_buffer;
    int work_number;
    MPI_Status status;
    
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    
    work_number = base;
    

    
    //Seed the slaves; send one unit of work to each slave.
    if (DEBUG) printf("... Seed the slaves; send one unit of work to each slave.\n");
    int proc;//proc 0 is the MASTER!
    for (proc = 1; proc<numprocs; proc++)
    {
        if (work_number == tope)
            break;
        
        output_buffer.work_number = work_number;
        output_buffer.size_msg = pob->genome[work_number].size;
        for(i=0; i<output_buffer.size_msg; i++){
            output_buffer.genes[i] = pob->genome[work_number].genes[i];
            output_buffer.identity[i] = identity[i];
        }
        output_buffer.flag=1;
        
        //pack_send_cromosoma(output_buffer, proc);
        MPI_Send(&output_buffer,
                 sizeof(msgGenome),
                 MPI_BYTE,
                 proc, //destination process
                 WORKTAG,
                 MPI_COMM_WORLD);
        //if (DEBUG) printf("processor MASTER %d  sent %d\n",myid,output_buffer);
        
        work_number++;
        
    }
    
    
    // Loop over getting new work requests until there is no more work
    //to be done
    if (DEBUG) printf("... Loop over getting new work requests\n");
    
    while (work_number < tope)
    {
        //Receive results from a slave
        MPI_Recv(&input_buffer,
                 sizeof(msgBusca),
                 MPI_BYTE,
                 MPI_ANY_SOURCE, //from any slave ID
                 MPI_ANY_TAG,
                 MPI_COMM_WORLD,
                 &status);
       
        pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
        
        //Send the slave a new work unit
        output_buffer.work_number = work_number;
        output_buffer.size_msg = pob->genome[work_number].size;
        for(i=0; i<output_buffer.size_msg; i++){
            output_buffer.genes[i] = pob->genome[work_number].genes[i];
            output_buffer.identity[i] = identity[i];
        }
        output_buffer.flag=1;
        
        MPI_Send(&output_buffer,
                 sizeof(msgGenome),
                 MPI_BYTE,
                 status.MPI_SOURCE, //to slave ID, who sent last message
                 WORKTAG,
                 MPI_COMM_WORLD);
        
        //if (DEBUG) printf("processor MASTER %d  sent %d\n",myid,output_buffer);
        
        work_number++;
    }//end while
    
    //There's no more work to be done, so receive all the outstanding
    //results from the slaves.
    if (DEBUG) printf("... There's no more work to be done, so receive all\n");
    for (proc = 1; proc<numprocs; proc++)
    {
        MPI_Recv(&input_buffer,
                 sizeof(msgBusca),
                 MPI_BYTE,
                 MPI_ANY_SOURCE, //from any slave ID
                 MPI_ANY_TAG,
                 MPI_COMM_WORLD,
                 &status);
        //if (DEBUG) printf("Processor MASTER %d  got %d\n",myid,input_buffer.result);
        pob->genome[input_buffer.index_pop].fitness = input_buffer.result;
    }
    
}

void slave(void){
    
    
   // printf("estou no escravo \n");
    
    
    int myid,i;
    msgGenome input_buffer;
    MPI_Status status;
    
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    
    permutation perm;//permutacion
    read_permutation_of_archivo("GENOME", &perm);// leitura validada
    is_cotail(perm.pi,identity,perm.size, size);
   // printf("entrei nesse seta aqui e sou o processo %d *********************** \n",myid);
    
    //function of DadosAdicionais.c
    set_TAMANHO_CAPPEPLUS( perm.pi, perm.size);
   // printf("qual o tamanho do TAMANHO_CAPPEPLUS %d\n",TAMANHO_CAPPEPLUS);
    //function of Preprocessamento.c
    load_nodeToCompMap();
    
    while(1){
        //Receive a message from the master
        MPI_Recv(&input_buffer,
                 sizeof(msgGenome),
                 MPI_BYTE,
                 0, //Master ID
                 MPI_ANY_TAG,
                 MPI_COMM_WORLD,
                 &status);
        
        //Check the tag of the received message.
        if (status.MPI_TAG == DIETAG)
            return;
        
        //calcula apenas o fitness
        if (input_buffer.flag)
        {
            msgBusca output_buffer;
            output_buffer.index_pop = input_buffer.work_number;// work(index)
            output_buffer.result = calcular_fitness_mpi(input_buffer); //do_work(work);
            //Send the result back
            MPI_Send(&output_buffer,
                 sizeof(msgBusca),
                 MPI_BYTE,
                 0, //Master ID
                 WORKTAG,
                 MPI_COMM_WORLD);
    
        }
        
        else//calcula busca local
        {
           // printf("entrei no els do calcula busca local \n");
            //printf("*****Recebi %d\n",input_buffer.work_number);
            msgBusca output_buffer;
           // output_buffer.index_pop = input_buffer.work_number;// work(index)
            int * result= auxlocal_search( input_buffer);
            output_buffer.index_gene = result[0];
            output_buffer.result = result[1];
            output_buffer.index_pop = result[2];
    
            
            free(result);
            //Send the result back
            MPI_Send(&output_buffer,
                     sizeof(msgBusca),
                     MPI_BYTE,
                     0, //Master ID
                     TAGLS,
                     MPI_COMM_WORLD);
            //printf("sai aqui??? \n");
            
        }//end else
        
        
        
    }//end while
}//end slave


void gerar_sementes(int tamanhoc,population *pob, int * sementes)
{
    int i,r,t,k,pos_gen;
    for (i=0; i<tamanhoc; i++) {
        while(1 && (qtd_genes/qtd_cromossomo>2))
		{
            
            r=rand();
            
            pos_gen = r % pob->genome[0].size;//values between 0 and (size)-1
            
            //printf("gerei ßß %d %d %d %d\n",pos_gen ,pob->genome[i].genes[pos_gen-1], pob->genome[i].genes[pos_gen+1],pob->genome[i].genes[pos_gen]);
            if((pos_gen!=0) && (pos_gen != pob->genome[0].size-1) && (pob->genome[0].genes[pos_gen-1]!=0)&&
               (pob->genome[0].genes[pos_gen+1]!=0) && (pob->genome[0].genes[pos_gen]!=0))
            {
             // printf(" infeliz i %d pos_gen %d\n",i, pos_gen);
                sementes[i]=pos_gen;
                break;
            }
  		}
    }

}

void master(void){
    
    /*Parametros*/
	float prob_crossover= 0.98;     //Crossover Probability
    float porc_selection=0.8;     //% of the population that will be selected for Crossover
    float porc_reemp=1.0 - 0.7;         //Pct. of the population from where it will be performed
    float prob_mutation=0.018;      //Mutation Probability
    float limiteentropia=0.3;         // % of reset population used in the calc of the entropy
    float porc_localsearch=0.6;   //% of the population that will be selected
    float porc_retain=0.6;        // % of the Preservation of the population when it is restarted
    float probability_ls;
    beta=0;
	/*Variables*/
	int i=0, num_generation;
	permutation perm;//permutacion
	population pob;//poblacion
	
    read_permutation_of_archivo("GENOME", &perm);// leitura validada
    int best_solution=perm.size;//cromosoma mejor_solucion;//mejor solucion encontrada
    
    if (is_cotail(perm.pi,identity,perm.size, size))
    {
        
        //function of DadosAdicionais.c
        set_TAMANHO_CAPPEPLUS( perm.pi, perm.size);
        //function of Preprocessamento.c
        load_nodeToCompMap();
        population_init(&pob, &perm, porc_reemp,  porc_selection);
        fitness_mpi(&pob, pob.basePopulation, pob.topPopulation);
        
        int * sementes=(int *) calloc (((pob.sizeTotal*2*perm.size)+20), sizeof (int));
       
        
        
        
        //gerar_sementes(((pob.sizeTotal*2*perm.size)+20), &pob, sementes);
        int beta1=0;
        
        /*
        for (i=0; i < 2271;i++) {
            
            printf("%d\n",sementes[beta1]);
             beta1++;
             printf("%d\n",sementes[beta1]);
             beta1++;
        }
         */
    
        local_search(&pob, 1,probability_ls,porc_localsearch,sementes);
        best_solution = pob.genome[0].fitness;
        num_generation = perm.size;
        for(i=2; i<=num_generation+1; i++)
        {
            if (i==3) {
                break;
            }
            
            selection(&pob,i,&best_solution);//seleccion:ordena populacao
            
	        crossover(&pob, &perm,prob_crossover);
            
            mutation(&pob, &perm,prob_mutation);
            
            fitness_mpi(&pob, pob.baseDescendencia, pob.topeDescendencia);
            
            replacement(&pob);
            
            probability_ls=(1.0-(i*1.0 /num_generation));
            local_search(&pob, i,probability_ls,porc_localsearch,sementes);
            
            if (reset_population(&pob,i,limiteentropia)){
                restarting_population(&pob, &perm, porc_retain);
            }
            
            
            
        }
        
        see_fitness( &pob ,0);
        selection(&pob,i,&best_solution);//seleccion:ordena populacao
       
        
        printf("Distance MAF %d\n", best_solution);
       
        
     
        for(i=0;i<pob.sizeTotal;i++){
            free(pob.genome[i].genes);
        }
         
        free(pob.genome);
       
    }
    else
        printf("***************** INPUT NOT IS-COTAIL ************************");
    

}




