/*
 ============================================================================
 Project     : Parallel Memetic Algorithm with Multi Populations no
 exchange individuals
 Author      : Lucas Angelo Silveira
 Universidade de Brasilia (UnB) - Brazil
 Group of Theory of Computation
 Version     : 1.0
 
 Last Modified On: February 01, 2016.
 ============================================================================
 */




#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "math.h"
#include "structure.h"
#include "MAMPNE.h"
#include "../Fitness/Processing.c"
#include "Calc_entropy.h"
#include "mpi.h"

float prob_crossover= 0.98;     //Crossover Probability
float porc_selection=0.8;     //% of the population that will be selected for Crossover
float porc_reemp=1.0 - 0.7;         //Pct. of the population from where it will be performed
float prob_mutation=0.018;      //Mutation Probability
//adjust parameter
float limiteentropia;         // % of reset population used in the calc of the entropy

//adjust this parameter
float porc_localsearch;   //% of the population that will be selected

float porc_retain;        // % of the Preservation of the population when it is restarted
int * identity;
int size;
int sentinela;


void read_permutation_of_archivo(char* name, permutation *perm)
{
    int i,n;
    FILE *filePointer;
    if ((filePointer = fopen(name,"r")) == NULL){
        printf("Atention file not finded\n");
        exit(1);
    }
    
    else
    {
        //read size of permutation
        fscanf(filePointer,"%d",&n);
        perm->size = n;
        size=n;
        //alocate memory for perm->pi
        perm->pi = malloc(perm->size*sizeof(int));
        //leer permutacion
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
        fscanf(filePointer,"%d",&n);
        identity= (int *) calloc (n, sizeof (int));
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&identity[i]);
        }
        
        
    }//end  else
    fclose(filePointer);
}//end read permutation of archivo



// Procedure that generates the initial population
void population_init(population *pob, permutation *perm){
	int i,j,r,n,prob;
	
    //Calc size of population
	n = perm->size;
    //printf("n%d\n",n);
	pob->sizePopulation =  (int) (n * (log(n) / log(2))); //siz of population = (nlogn); log in base 2
    // printf(" quanto tenho de log %d\n ",(int)(n * (log(n) / log(2))));
	pob->basePopulation = 0;
	pob->topPopulation = pob->sizePopulation;
    
	pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de population what go to crossover
	pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a partir de where will replace
    
	pob->baseDescendencia = pob->sizePopulation;
	pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
	pob->sizeTotal = pob->topeDescendencia;// tamanho total = poblacion + descendencia
    
    /*alocar memoria*/
	pob->genome = malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in population_init(x,x)\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = malloc(n * sizeof(int));//alocar memoria para cada genome
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in population_init(x,x)\n");
    }
    
	/*create population a partir of  permutacion*/
    // A genome is see with a permutation of concatenation chroms
    
	for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        
		for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//valores entre 0 y 1
                if (prob == 1)//si prob 1, entonces gen positivo
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                }
                else//si prob 0, entonces gen negativo
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                    
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
                
            }
        }//end for interno
        
		pob->genome[i].size = perm->size;
        //eliminar_repetidas(&pob,pob->sizePopulation,adicional,n);
	}//end for extern
	/*agregar tamanho a desc*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
		pob->genome[i].size = perm->size;
	}
    
    
}//end create init population

//compute the distance to each genome of the population
void fitness_population(population *pob, int flag){
	int i;
    
	if (flag > 1)
	{ //fitness sobre descendencia
        
        
        
		for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
		{
            
            
            /*  if (flag==8) {
             int j;
             
             for (j=0; j< pob->genome[i].size;j++) {
             printf("%d,",pob->genome[i].genes[j]);
             }
             
             printf("\n");
             }
             */
		 	load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            pob->genome[i].fitness =distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            
            
        }
        
	}
	else
    { // fitness sobre poblacion
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
        {
            /*
             int j;
             
             for (j=0; j< pob->genome[i].size;j++) {
             printf("%d,",pob->genome[i].genes[j]);
             }
             
             printf("\n");
             */
            load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            pob->genome[i].fitness = distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            //   printf("passei \n");
            
        }
    }
}


void countingSort(population *pob, int k){
	int i,j;
	int *c;//c[k];
	Genome *a;
    
    /*alocar memoria para c[k]*/
    c = malloc(k*sizeof(int));
    
	a = malloc(pob->sizePopulation*sizeof(Genome));
	if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
	for(i=0; i<pob->sizePopulation; i++){
		a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
	}
    
	/*counting sort*/
	for(i=0; i<k; i++){
		c[i] = 0;
	}
    
	for(i=0; i<pob->sizePopulation; i++){
		c[a[i].fitness] = c[a[i].fitness] + 1;
	}
    
	for(i=1; i<k; i++){
		c[i] = c[i] + c[i-1];
	}
    
	for(i=pob->sizePopulation-1; i>=0; i--){
		
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
		c[a[i].fitness] = c[a[i].fitness] - 1;
	}
    
    
    /*Liberar memoria*/
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
	free(a);
    free(c);
}


void selection(population *pob,  int n_generetion, int * best_solution){
	
    countingSort(pob,pob->genome[0].size+2);
    
    if (pob->genome[0].fitness < *best_solution)
        *best_solution = pob->genome[0].fitness;
    
}


void crossover(population *pob, permutation *perm){
	int i,j,k,pos_pai1, pos_pai2, point_crossover;
	float prob;
	k = pob->baseDescendencia;//indice para colocar en descendencia
	/*crossver sobre populacao atual*/
	for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
		//se falta ultima posicao para colocar descendencia, terminar
		if (k+1 == pob->sizeTotal)
			break;
        
		//escolher posicao dos pais
		pos_pai1 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		//buscar pos2 que no seja igual a pos1, pelo  menos 10vezes
		for(j=0; j<10; j++){
			if (pos_pai1 != pos_pai2)
                break;
			pos_pai2 = rand() % pob->limiteSelection;//valores entre 0 e limiteSelection-1
		}
        
		prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
		if (prob <= prob_crossover)
        {//crossover
			point_crossover = rand() % perm->size;//valores entre 0 e tamanho de permutation - 1
			//aplicar single_point_crossover na primera parte antes do ponto de crossover
			for(j=0; j<point_crossover; j++){
				pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			//aplicar single_point_crossover na segunda parte depois do ponto de crossover
			for(j=point_crossover; j<perm->size; j++){
				pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
				pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
			}
			k = k + 2;
		}//end if crossover
	}//end for iniit
	pob->topeDescendencia = k;//atualizar tope descendencia
}


void mutation(population *pob, permutation *perm){
	int i,j;
	float prob;
    
	/*mutation about descendents, baseDescendencia=0*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		for(j=0; j<perm->size; j++)
        {
			prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
			
            if(prob <= prob_mutation)
            {//applied a gene mutation (flip sign)
                
                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
                
            }
            
        }
        
    }//end for
    
}//end of the function mutation

void replacement(population *pob){
	int i,j,pos_reemp;
    int conta=0;
	/*substituir descendentes na populacao atual*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		//buscar pos de substituicao
		int tamanhoReemp = pob->topPopulation - pob->baseReplace;
		pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(valores entre 0 e tamanhoReemp-1)+baseReplace
		
        //realizar substituicao
		if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
			//copiar  fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copiar genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
	}
}

/* operator local search based in mutation*/
/* The operator is applied over one individual "i"
 * Return 1(true) when we have better results of the indidual i"
 * otherwise return 0(false) */

int auxlocal_search(population *pob, int i){
	int k, r, replace, fitness, pos_gen, better;
    
	// nro of iteration (the genes) maximum the be modified
	k=0;
	replace = 0;//false
	better = pob->genome[i].fitness;
    pos_gen=0;
	while(k<2)
    {
        
		while(1 && (qtd_genes/qtd_cromossomo>2))
		{
            
            r=rand();
            pos_gen = r % pob->genome[0].size;//values between 0 and (size)-1
            if((pos_gen!=0) && (pos_gen != pob->genome[0].size-1) && (pob->genome[i].genes[pos_gen-1]!=0)&&
               (pob->genome[i].genes[pos_gen+1]!=0) && (pob->genome[i].genes[pos_gen]!=0))
                break;
            
        }
        
        if (pos_gen)
        {
            
            
            pob->genome[i].genes[pos_gen] = pob->genome[i].genes[pos_gen]*(-1);
            
	        load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            
	        fitness =distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            if (fitness < better){
                pob->genome[i].fitness = fitness;
                replace = 1;//true
                break;
                
            }
            else{
                pob->genome[i].genes[pos_gen] =pob->genome[i].genes[pos_gen]*(-1);//restore previous state
            }
        }
        k++;
	}//end while
    
	return replace;
}


/* Procedure that applies Local Search over the population
 * if nro_generacion is equal to 1, it means that is the initial pop0
 */
void local_search(population *pob, int nro_generation){
    int i;
	int melhorou=0;
    if (nro_generation > 1)
    {
        /* limit until where we apply local search */
        int limit_search = pob->sizeTotal * porc_localsearch;
        /*local search over a set of the population*/
        for(i=pob->basePopulation; i < limit_search; i++)
        {
            if(auxlocal_search(pob, i))
                melhorou++;
        }
    }
    else
    { //nro_generacion == 1, we have initial population
        for(i=pob->basePopulation; i < pob->topPopulation; i++)
            auxlocal_search(pob, i);
        
    }
    
}

/* method that resert the population, conservando una parte
 * e gerando aleatoriamente o resto
 */
void restarting_population(population *pob, permutation *perm){
	int i,j, prob,r, size_conservation;
    
	/* retain a portion of the population */
	size_conservation = pob->sizePopulation * porc_retain;
    
	for(i=size_conservation; i < pob->sizePopulation; i++)
    {
		for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//valores entre 0 y 1
                if (prob == 1)
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                    
                }
                else//si prob 0, entonces gen negativo
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
            }
        }//end for interno
        
		pob->genome[i].size = perm->size;
        
        //calculate translocation distance for each genome
        load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
        pob->genome[i].fitness = distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
        auxlocal_search(pob, i);
        
	}
    
}//end restarting_population



//this function return 1 for reset population and 0 otherwise
int reset_population(population *pob, int i){
    float value=entropia_shannon(pob);
    if ( value < limiteentropia)
    {
        return 1;
        
    }
    
    else
        return 0;
    
}

void see_population(permutation *perm,population *pob)
{
    int i,j;
    for(i=0; i<pob->sizePopulation;i++ )
    {
        printf("%d ",i);
        for(j=0;j<perm->size;j++)
        {
            printf("%d,", pob->genome[i].genes[j]);
            
        }
        printf("\n");
    }
}//end see_population


void see_fitness(population *pob ,int flag)
{
    int i;
    if(!flag)
    {
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
            printf("%d: distance is:%d\n",i, pob->genome[i].fitness);
    }
    else
    {
        
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
            printf("distance is:%d\n",pob->genome[i].fitness);
    }
    
}





int * allocate_space(){
int * a = (int *) calloc(2, sizeof (int));
    return a;
}

void my_bcast( int geration, MPI_Comm communicator, int world_rank, int flag, population *pob,float prob_insertion) {
    int world_size, j,ii, iii,i;
    //float prob_insertion=0.7;
    MPI_Comm_size(communicator, &world_size);
    msgGenome enviar,recebe;
    
    // If we are the root process, send our data to everyone
    
    enviar.size=pob->genome[0].size;
    enviar.fitness=pob->genome[0].fitness;
    for (i=0; i<pob->genome[0].size; i++) {
        enviar.genes[i]=pob->genome[0].genes[i];
    }
    
    for (ii = 0; ii < world_size; ii++)
    {
        if (ii != world_rank)
        {
            int error_code= MPI_Send(&enviar,
                                        sizeof(msgGenome),
                                        MPI_BYTE,
                                        ii,//destination process
                                        0,
                                        communicator);
              
            if (error_code!=MPI_SUCCESS)
            {
                char error_string[BUFSIZ];
                int length_of_error_string;
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                }
            
            }//end if
            
            else
            {
                for( iii = 0; iii < world_size; iii++)
                {
                    
                    if (iii != world_rank)
                    {
                        int error_code=MPI_Recv(&recebe,
                                                sizeof(msgGenome),
                                                MPI_BYTE,
                                                iii, // process that sended
                                                0,
                                                communicator,
                                                MPI_STATUS_IGNORE);
                        
                        if (error_code != MPI_SUCCESS)
                        {
                            char error_string[BUFSIZ];
                            int length_of_error_string;
                            MPI_Error_string(error_code, error_string, &length_of_error_string);
                        }//end if
                        
			
                        if (flag)
                        {
                            
                           
                                int controle=0;
                                for (j=0; j<pob->sizePopulation; j++) {
                                    if (recebe.fitness < pob->genome[j].fitness)
                                    {
                                        pob->genome[j].fitness=recebe.fitness;
                                        for (i=0; i<pob->genome[j].size; i++)
                                        {
                                            pob->genome[j].genes[i]=recebe.genes[i];
                                        }//end for
                                        
                                        controle=1;
                                    }//end if
                                    
                                    if (controle)
                                        break;
                                    
                                    
                            }//end for
                            
                            
                        }//insertion in the population
                        
                    }//end if
                }//end for
                
            }//else
        }
    
}

void MemetiAlg(MPI_Comm communicator, int world_rank, float parameter){
    int i=0;
    permutation perm;
    sentinela=0;
    population pob;
    int best_solution,num_generation;
    read_permutation_of_archivo("GENOME", &perm);
    
    //check if is-cotail
    if (is_cotail(perm.pi,identity,perm.size, size))
    {
        limiteentropia=0.3;
        porc_localsearch=0.6;
        porc_retain=0.6;
 	    parameter=0.2;
        population_init(&pob, &perm);
        
        //function of DadosAdicionais.c
        set_TAMANHO_CAPPEPLUS( perm.pi, perm.size);
        //function of Preprocessamento.c
        load_nodeToCompMap();
        
        fitness_population(&pob, 1);
        local_search(& pob, 1);
        best_solution = pob.genome[0].fitness;
        
        num_generation = perm.size;
        for(i=2; i<=num_generation+1; i++)
        {
            selection(&pob,i,&best_solution);
            crossover(&pob, &perm);
            mutation(&pob, &perm);
            fitness_population(&pob,i);
            replacement(&pob);
            local_search(&pob, num_generation);
            if (reset_population(&pob,i)){
                restarting_population(&pob, &perm);
            }
            
            
            countingSort(&pob,(pob.genome[0].size+2));
            if (i==(num_generation+1))
            {
                my_bcast( i, communicator, world_rank,1,&pob,parameter);
            }
            MPI_Barrier(communicator);
        }//end for
        
        selection(&pob,i,&best_solution);//seleccion:ordena populacao
        
        if (world_rank==0) {
            printf("Distance MAMPNE:%d, ", best_solution);
            //see_fitness(&pob, 0);
        }
        
    
    }
    else
        printf("***************** INPUT NOT IS-COTAIL ************************");
    
}


